// file      : xsd/cxx/parser/parser-header.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/parser-header.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      struct List: Traversal::List, protected virtual Context
      {
        List (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          String const& name (ename (l));
          SemanticGraph::Type& t (l.argumented ().type ());
          String type_name (ename (t));

          //@@ What's to come is a nasty hack.
          //
          String item_name;

          if (t.is_a<SemanticGraph::Fundamental::IdRef> ())
          {
            item_name = L"IDREF" == name ? L"IDREF_" : L"IDREF";
          }
          else if (t.is_a<SemanticGraph::Fundamental::IdRefs> ())
          {
            item_name = L"IDREFS" == name ? L"IDREFS_" : L"IDREFS";
          }
          else
            item_name = type_name == name ? type_name + L"_" : type_name;

          os << "namespace _xsd_" << name
             << "{"
             << "template < typename _xsd_" << item_name << " >" << endl
             << "struct " << name << ": virtual ::xsd::cxx::parser::" <<
            (validation ? "validating" : "non_validating") << "::list< " <<
            "_xsd_" << item_name << ", " << char_type << " >"
             << "{"
             << "};"
             << "}";

          os << "template < typename _xsd_" << name << "_, "
             << "typename _xsd_" << item_name << " >" << endl
             << "struct " << name << ":" << endl
             << "virtual " << parser_template << "< " <<
            "_xsd_" << name << "_, " << char_type << " >," << endl
             << "virtual _xsd_" << name << "::" << name << "< " <<
            "_xsd_" << item_name << " >"
             << "{"
             << "/*" << endl
             << endl
             << "Parser hooks. Override them in your implementation." << endl
             << endl
             << "virtual void pre ();"
             << endl
             << "virtual void item (const _xsd_" << item_name << "&);"
             << endl
             << "virtual _xsd_" << name << "_ post ();"
             << endl
             << "*/"
             << "};";
        }
      };


      struct Union: Traversal::Union, protected virtual Context
      {
        Union (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          String const& name (ename (u));

          os << "template < typename _xsd_" << name << "_ >" << endl
             << "struct " << name <<
            ": virtual " << parser_template << "< _xsd_" << name << "_, " <<
            char_type << " >"
             << "{"
             << "/*" << endl
             << endl
             << "Parser hooks. Override them in your implementation." << endl
             << endl
             << "virtual void pre ();"
             << endl
             << "virtual void _characters (const " << string_type << "&);"
             << endl
             << "virtual _xsd_" << name << "_ post ();"
             << endl
             << "*/"
             << "};";
        }
      };


      //
      //
      struct ParserHook: Traversal::Member,
                         protected virtual Context
      {
        ParserHook (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          String type (L"_xsd_" + name + L"_");

          os << "virtual void " << name << " (const " << type << "&);";
        }
      };


      //
      //
      struct ParserAccessor: Traversal::Member,
                             protected virtual Context
      {
        ParserAccessor (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          String type (L"_xsd_" + name + L"_");

          os << "void" << endl
             << name << "_parser (" << parser_template << "< " <<
            type << ", " << char_type << " >&);"
             << endl;
        }
      };


      //
      //
      struct ParserMember: Traversal::Member,
                           protected virtual Context
      {
        ParserMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          String type (L"_xsd_" + name + L"_");
          String inst (L"_xsd_" + name + L"__");

          os << parser_template << "< " << type << ", " <<
            char_type << " >* " << inst << ";";
        }
      };


      //
      //
      struct MemberStruct: Traversal::Member,
                           protected virtual Context
      {
        MemberStruct (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          String targ (unclash (name, "x"));

          os << "template < typename " << targ << " >" << endl
             << "struct _xsd_" << name
             << "{"
             << "virtual ~_xsd_" << name << " ();"
             << "virtual void " << name << " (const " << targ << "&);"
             << "void _xsd_call_glue_" << name << " (" <<
            parser_template << "< " << targ << ", " << char_type << " >*);"
             << "};";
        }
      };

      //
      //
      struct MemberStructBase: Traversal::Member,
                               protected virtual Context
      {
        MemberStructBase (Context& c, String type_name)
            : Context (c), type_name_ (type_name)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));

          os << "," << endl
             << "  _xsd_" << name << "< _xsd_" << name << "_ >";
        }

      private:
        String type_name_;
      };

      //
      //
      struct Particle: Traversal::All,
                       Traversal::Choice,
                       Traversal::Sequence,
                       protected virtual Context
      {
        Particle (Context& c)
            : Context (c)
        {
          *this >> contains_particle_ >> *this;
        }


        virtual Void
        traverse (SemanticGraph::All& a)
        {
          if (!a.context().count ("comp-number"))
            return;

          UnsignedLong state_count (
            a.context().get<UnsignedLong> ("state-count"));

          os << "void" << endl
             << "all_0 (unsigned long& state," << endl
             << "unsigned char* count," << endl
             << "const " << string_type << "& ns," << endl
             << "const " << string_type << "& n," << endl
             << "bool start);"
             << endl
             << "::xsd::cxx::parser::validating::all_count< " <<
            state_count << "UL > v_all_count_;"
             << endl;
        }

        virtual Void
        traverse (SemanticGraph::Choice& c)
        {
          if (!c.context().count ("comp-number"))
            return;

          UnsignedLong n (c.context ().get<UnsignedLong> ("comp-number"));

          os << "void" << endl
             << "choice_" << n << " (unsigned long& state," << endl
             << "unsigned long& count," << endl
             << "const " << string_type << "& ns," << endl
             << "const " << string_type << "& n," << endl
             << "bool start);"
             << endl;

          Traversal::Choice::traverse (c);
        }

        virtual Void
        traverse (SemanticGraph::Sequence& s)
        {
          if (!s.context().count ("comp-number"))
            return;

          UnsignedLong n (s.context ().get<UnsignedLong> ("comp-number"));

          os << "void" << endl
             << "sequence_" << n << " (unsigned long& state," << endl
             << "unsigned long& count," << endl
             << "const " << string_type << "& ns," << endl
             << "const " << string_type << "& n," << endl
             << "bool start);"
             << endl;

          Traversal::Sequence::traverse (s);
        }

      private:
        Traversal::ContainsParticle contains_particle_;
      };

      //
      //
      struct Complex : Traversal::Complex,
                       protected virtual Context
      {
        Complex (Context& c)
            : Context (c),
              parser_hook_ (c),
              parser_member_ (c),
              parser_accessor_ (c),
              member_struct_ (c)
        {
          names_parser_hook_ >> parser_hook_;
          names_parser_member_ >> parser_member_;
          names_parser_accessor_ >> parser_accessor_;
          names_member_struct_ >> member_struct_;
        }

        virtual Void
        traverse (Type& c)
        {
          String const& name (ename (c));
          String type (L"_xsd_" + name + L"_");

          // In case of an inheritance-by-restriction, we don't need to
          // generate parser hooks, etc. since they are the same as in
          // the base. We only need the parsing/validation code.
          //
          Boolean restriction (restriction_p (c));
          Boolean he (has<Traversal::Element> (c));
          Boolean ha (has<Traversal::Attribute> (c));

          Boolean htp (false);
          {
            HasTemplateParam t (*this, htp);
            t.traverse (c);
          }

          //
          //
          os << "namespace _xsd_" << name
             << "{";

          if (htp)
          {
            if (!restriction)
              names (c, names_member_struct_);

            //
            //
            os << "template < ";

            {
              TemplateParamDeclList template_param_decl_list (*this);
              template_param_decl_list.traverse (c);
            }

            os << " >" << endl;

            os << "struct " << name
               << ":" << endl;

            {
              os << "  ";

              if (c.inherits_p ())
              {
                BaseType base (*this, name, true);
                Traversal::Inherits inherits_base (base);

                inherits (c, inherits_base);
              }
              else
                os << "virtual " << parser_base;

            }

            if (!restriction)
            {
              MemberStructBase base (*this, name);
              Traversal::Names names_base (base);

              names (c, names_base);
            }

            os << "{";

            //
            //
            if (!restriction && (he || ha))
            {
              os << "// Parser construction API." << endl
                 << "//" << endl;

              names (c, names_parser_accessor_);

              os << "void" << endl
                 << "parsers (";

              {
                ParserParamDecl decl (*this, false);
                decl.traverse (c);
              }

              os << ");"
                 << endl;

              // Default c-tor.
              //
              os << name << " ();"
                 << endl
                 << endl;
            }


            // Implementation.
            //
            if (he || ha)
            {
              os << "// Implementation." << endl
                 << "//" << endl
                 << "protected:" << endl;
            }

            if (he)
            {
              // _start_element_impl
              //
              os << "virtual bool" << endl
                 << "_start_element_impl (const " << string_type << "&," << endl
                 << "const " << string_type << "&);"
                 << endl;

              // end_element
              //
              os << "virtual bool" << endl
                 << "_end_element_impl (const " << string_type << "&," << endl
                 << "const " << string_type << "&);"
                 << endl;
            }

            if (ha)
            {
              // attribute
              //
              os << "virtual bool" << endl
                 << "_attribute_impl (const " << string_type << "&," << endl
                 << "const " << string_type << "&," << endl
                 << "const " << string_type << "&);"
                 << endl;
            }

            if (!restriction && (he || ha))
            {
              os << "protected:" << endl;
              names (c, names_parser_member_);
              os << endl;
            }

            if (validation && he)
            {
              os << "protected:" << endl

                 << "::xsd::cxx::parser::validating::state_stack< " <<
                name << ", " << char_type <<  " > v_state_stack_;"
                 << endl;

              os << "virtual void" << endl
                 << "_pre_impl ();"
                 << endl;

              os << "virtual void" << endl
                 << "_post_impl ();"
                 << endl;

              Particle t (*this);
              t.dispatch (c.contains_compositor ().compositor ());
            }

            os << "};";
          }
          else
          {
            os << "typedef ";

            if (c.inherits_p ())
            {
              BaseType base (*this, name, false);
              Traversal::Inherits inherits_base (base);

              inherits (c, inherits_base);
            }
            else
              os << parser_base;


            os << " " << name << ";";
          }

          os << "}"; // namespace


          // class name
          //
          os << "template < " <<
            "typename _xsd_" << name << "_";

          {
            TemplateParamDeclList template_param_decl_list (*this, false);
            template_param_decl_list.traverse (c);
          }

          os << " >" << endl;

          os << "struct " << name << ":" << endl
             << "  virtual " << parser_template << "< " <<
            "_xsd_" << name << "_, " <<
            char_type << " >," << endl
             << "  virtual _xsd_" << name << "::" << name;

          {
            TemplateParamUseList template_param_use_list (*this);
            template_param_use_list.traverse (c);
          }

          os << "{"
             << "/*" << endl
             << endl
             << "Parser hooks. Override them in your implementation." << endl
             << endl
             << "virtual void pre ();";


          if (!restriction && (ha || he))
          {
            os << endl;
            names (c, names_parser_hook_);
            os << endl;
          }

          os << "virtual _xsd_" << name << "_ post ();"
             << endl
             << "*/" << endl
             << endl
             << "};";
        }

      private:
        //
        //
        ParserHook parser_hook_;
        Traversal::Names names_parser_hook_;

        //
        //
        ParserMember parser_member_;
        Traversal::Names names_parser_member_;

        //
        //
        ParserAccessor parser_accessor_;
        Traversal::Names names_parser_accessor_;

        //
        //
        MemberStruct member_struct_;
        Traversal::Names names_member_struct_;
      };


      //
      //
      struct Includes : Traversal::Imports,
                        Traversal::Includes,
                        protected virtual Context
      {
        Includes (Context& c, String const& expr)
            : Context (c), expr_ (expr)
        {
        }

        virtual Void
        traverse (SemanticGraph::Imports& i)
        {
          traverse (i.path ());
        }

        virtual Void
        traverse (SemanticGraph::Includes& i)
        {
          traverse (i.path ());
        }

        virtual Void
        traverse (SemanticGraph::Path const& path)
        {
          using BackendElements::Regex::perl_s;

          // Note the use of the portable representation of the path.
          //
          String const& name (perl_s (String (path.string ()), expr_));

          Char op (include_with_brackets ? '<' : '"');
          Char cl (include_with_brackets ? '>' : '"');

          os << "#include " << op << include_prefix << name << cl << endl
             << endl;
        }

      private:
        String expr_;
      };

      struct FundamentalNamespace : Namespace,
                                    protected virtual Context
      {
        FundamentalNamespace (Context& c)
            : Context (c),
              Namespace (c)
        {
        }

        void
        traverse (Type& ns)
        {
          pre (ns);

          String impl ("::xsd::cxx::parser::");
          impl += (validation ? L"validating" : L"non_validating");

          os << "// Generic parser." << endl
             << "//" << endl;

          String g (L"XSD_CXX_PARSER_" + xml_parser + L"_" + char_type +
                    L"_PARSER_IN_" + ns_name (ns));

          std::transform (g.begin (), g.end(), g.begin (), upcase);
          g = escape (g); // Make it a C++ id.

          os << "#ifndef " << g << endl
             << "#define " << g << endl
             << endl
             << "template <typename X>" << endl
             << "struct parser: virtual " << parser_template << "< X, " <<
            char_type << " >," << endl
             << "virtual " << impl << "::empty_content< " << char_type << " >"
             << "{"
             << "};"
             << "template <typename X>" << endl
             << "struct document: ::xsd::cxx::parser::" << xml_parser <<
            "::document< X, " << char_type << " >"
             << "{"
             << "document (" << parser_template << "< X, " <<
            char_type << " >& p," << endl
             << "const " << string_type << "& root_element_name)" << endl
             << ": ::xsd::cxx::parser::" << xml_parser <<
            "::document< X, " << char_type << " > (" << endl
             << "p," << endl
             << "root_element_name)"
             << "{"
             << "}"
             << "document (" << parser_template << "< X, " <<
            char_type << " >& p," << endl
             << "const " << string_type << "& root_element_namespace," << endl
             << "const " << string_type << "& root_element_name)" << endl
             << ": ::xsd::cxx::parser::" << xml_parser <<
            "::document< X, " << char_type << " > (" << endl
             << "p," << endl
             << "root_element_namespace," << endl
             << "root_element_name)"
             << "{"
             << "}"
             << "};"
             << "#endif" << endl
             << endl
             << endl;

          os << "// Parsers for XML Schema built-in types." << endl
             << "//" << endl
             << endl;

          os << "// 8-bit" << endl
             << "//" << endl
             << "typedef " << impl << "::byte< " <<
            char_type << " > byte;"
             << "typedef " << impl << "::unsigned_byte< " <<
            char_type << " > unsigned_byte;"
             << endl;


          os << "// 16-bit" << endl
             << "//" << endl
             << "typedef " << impl << "::short_< " <<
            char_type << " > short_;"
             << "typedef " << impl << "::unsigned_short< " <<
            char_type << " > unsigned_short;"
             << endl;


          os << "// 32-bit" << endl
             << "//" << endl
             << "typedef " << impl << "::int_< " <<
            char_type << " > int_;"
             << "typedef " << impl << "::unsigned_int< " <<
            char_type << " > unsigned_int;"
             << endl;


          os << "// 64-bit" << endl
             << "//" << endl
             << "typedef " << impl << "::long_< " <<
            char_type << " > long_;"
             << "typedef " << impl << "::unsigned_long< " <<
            char_type << " > unsigned_long;"
             << endl;

          os << "// Boolean." << endl
             << "//" << endl
             << "typedef " << impl << "::boolean< " <<
            char_type << " > boolean;"
             << endl;

          os << "// Floating-point types." << endl
             << "//" << endl
             << "typedef " << impl << "::float_< " <<
            char_type << " > float_;"
             << "typedef " << impl << "::double_< " <<
            char_type << " > double_;"
             << "typedef " << impl << "::decimal< " <<
            char_type << " > decimal;"
             << endl;

          os << "// String." << endl
             << "//" << endl
             << "typedef " << impl << "::string< " <<
            char_type << " > string;"
             << endl
             << endl;


          os << "// Exceptions. See libxsd/xsd/cxx/parser/exceptions.hxx " <<
            "for details." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::parser::exception< " <<
            char_type << " > exception;"
             << "typedef ::xsd::cxx::parser::parsing< " <<
            char_type << " > parsing;"
             << "typedef ::xsd::cxx::parser::unexpected_element< " <<
            char_type << " > unexpected_element;"
             << endl
             << endl;

          if (xml_parser == L"xerces")
          {
            os << "// Parsing flags. See " <<
              "libxsd/xsd/cxx/parser/elements.hxx for details." << endl
               << "//" << endl
               << "typedef ::xsd::cxx::parser::xerces::flags flags;"
               << endl
               << endl;

            os << "// Parsing properties. See " <<
              "libxsd/xsd/cxx/parser/elements.hxx for details." << endl
               << "//" << endl
               << "typedef ::xsd::cxx::parser::xerces::properties< " <<
              char_type << " > properties;"
               << endl
               << endl;
          }

          if (xml_parser == L"expat")
          {
            os << "// Simple auto pointer for Expat's XML_Parser object." << endl
               << "//" << endl
               << "typedef ::xsd::cxx::parser::expat::parser_auto_ptr " <<
              "parser_auto_ptr;"
               << endl
               << endl;
          }

          post (ns);
        }
      };
    }

    Void
    generate_parser_header (Context& ctx,
                            SemanticGraph::Schema& tu,
                            String const& expr)
    {
      String c (ctx.char_type);

      // At the moment we only support UTF-8 expat.
      //
      if (ctx.xml_parser == L"expat")
      {
        ctx.os << "#ifdef XML_UNICODE" << endl
               << "#error UTF-16 expat (XML_UNICODE defined) is not " <<
          "yet supported" << endl
               << "#endif" << endl
               << endl;
      }

      //
      //
      if (c == L"char")
      {
        ctx.os << "#ifndef XSD_USE_CHAR" << endl
               << "#define XSD_USE_CHAR" << endl
               << "#endif" << endl
               << endl;

        ctx.os << "#ifndef XSD_CXX_PARSER_USE_CHAR" << endl
               << "#define XSD_CXX_PARSER_USE_CHAR" << endl
               << "#endif" << endl
               << endl;
      }
      else if (c == L"wchar_t")
      {
        ctx.os << "#ifndef XSD_USE_WCHAR" << endl
               << "#define XSD_USE_WCHAR" << endl
               << "#endif" << endl
               << endl;

        ctx.os << "#ifndef XSD_CXX_PARSER_USE_WCHAR" << endl
               << "#define XSD_CXX_PARSER_USE_WCHAR" << endl
               << "#endif" << endl
               << endl;
      }

      ctx.os << "#include <xsd/cxx/parser/exceptions.hxx>" << endl
             << "#include <xsd/cxx/parser/elements.hxx>" << endl;

      if (ctx.validation)
        ctx.os << "#include <xsd/cxx/parser/validating/parser.hxx>" << endl
               << "#include <xsd/cxx/parser/validating/types.hxx>" << endl;
      else
        ctx.os << "#include <xsd/cxx/parser/non-validating/parser.hxx>" << endl
               << "#include <xsd/cxx/parser/non-validating/types.hxx>" << endl;

      ctx.os << "#include <xsd/cxx/parser/" << ctx.xml_parser <<
        "/elements.hxx>" << endl
             << endl;

      // Emit fundamental types.
      //
      {
        Traversal::Schema schema, xsd;
        Traversal::Implies implies;
        Traversal::Names names;
        FundamentalNamespace ns (ctx);

        schema >> implies >> xsd >> names >> ns;

        schema.dispatch (tu);
      }

      Traversal::Schema schema;

      Traversal::Sources sources;
      Includes includes (ctx, expr);
      Traversal::Names schema_names;

      Namespace ns (ctx);
      Traversal::Names names;

      schema >> includes;
      schema >> sources >> schema;
      schema >> schema_names >> ns >> names;

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);

      names >> list;
      names >> union_;
      names >> complex;

      schema.dispatch (tu);
    }
  }
}

