// file      : xsd/cxx/tree/stream-extraction-template.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/stream-extraction-template.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c, String const& name_ = "")
            : Context (c), name (name_)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
            name = ename (l);

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "template <typename _istream>" << endl
             << scoped_name << "::" << endl
             << name << " (::xsd::cxx::tree::istream< _istream >& s," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ::xsd::cxx::tree::simple_type (f, c)," << endl
             << "::xsd::cxx::tree::sequence< " <<
            item_type_name (l.argumented ().type ()) << " > (s, f, c)"
             << "{"
             << "}";
        }

      private:
        String
        item_type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

      private:
        String name;
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c, String const& name_ = "")
            : Context (c),
              name (name_)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
            name = ename (u);

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          String base (L"::xsd::cxx::tree::string< " + char_type + L" >");

          os << "template <typename _istream>" << endl
             << scoped_name << "::" << endl
             << name << " (::xsd::cxx::tree::istream< _istream >& s," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": " << base << " (s, f, c)"
             << "{"
             << "}";
        }

      private:
        String name;
      };


      struct Enumeration : Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c, String const& name_ = "")
            : Context (c),
              base_ (c),
              name (name_)
        {
          inherits_base_ >> base_;
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
            name = ename (e);

          Boolean string_based (false);
          {
            IsStringBasedType t (string_based);
            t.dispatch (e);
          }

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          os << "template <typename _istream>" << endl
             << scoped_name << "::" << endl
             << name << " (::xsd::cxx::tree::istream< _istream >& s," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ";

          inherits (e, inherits_base_);

          os << " (s, f, c)"
             << "{";

          if (string_based)
            os << "_xsd_" << name << "_convert ();";

          os << "}";
        }

      private:
        Traversal::Inherits inherits_base_;
        BaseTypeName base_;

        String name;
      };


      //
      //
      template <typename Complex>
      struct Member : Traversal::Member, protected virtual Context
      {
        Member (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));
          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            Traversal::Belongs belongs_;

            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };


      struct CtorMember : Traversal::Member,
                          protected virtual Context
      {
        CtorMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (SemanticGraph::Member& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          /*
          if (max (m) != 1)
          {
            // sequence
            //
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
          }
          else
          {
            //
            // one
          }
          */

          os << "," << endl
             << "_xsd_" << name << "_ (s," << endl
             << "f | ::xsd::cxx::tree::flags::not_root," << endl
             << "this)";
        }
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c, String const& name_ = "")
            : Context (c),
              name (name_),
              member_ (c),
              ctor_member_ (c),
              base_ (c)
        {
          names_member_ >> member_;
          names_ctor_member_ >> ctor_member_;
          inherits_ >> base_;
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
            name = ename (c);

          enter_scope (name);
          String scoped_name (scope);
          leave_scope ();

          // Go after anonymous types first.
          //
          {
            enter_scope (name);
            names (c, names_member_);
            leave_scope ();
          }

          os << "template <typename _istream>" << endl
             << scoped_name << "::" << endl
             << name << " (::xsd::cxx::tree::istream< _istream >& s," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << "::xsd::cxx::tree::type* c)" << endl
             << ": ";

          if (c.inherits_p ())
            inherits (c, inherits_);
          else
            os << "::xsd::cxx::tree::type";

          os << " (s, f, c)";

          names (c, names_ctor_member_);

          os << "{"
             << "}";
        }

      private:
        String name;

        Member<Complex> member_;
        Traversal::Names names_member_;

        CtorMember ctor_member_;
        Traversal::Names names_ctor_member_;

        Traversal::Inherits inherits_;
        BaseTypeName base_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element,
                            GlobalElementBase,
                            protected virtual Context
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c), Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (!generate_p (e))
            return;

          SemanticGraph::Type& t (e.type ());

          if (!t.named ())
          {
            String const& name (ename (e));

            Traversal::Belongs belongs_;

            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (e, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };
    }

    Void
    generate_stream_extraction_template (Context& ctx,
                                         SemanticGraph::Schema& tu)
    {
      ctx.os << "#include <xsd/cxx/tree/stream-extraction.hxx>" << endl
             << endl;

      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names names_ns, names;

      Namespace ns (ctx, false);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;

      schema.dispatch (tu);
    }
  }
}
